from abc import ABC, abstractmethod
from telegrinder.api.abc import ABCAPI
from telegrinder.types import Update
from .view.abc import ABCView
import typing


class ABCDispatch(ABC):
    @abstractmethod
    def feed(self, event: Update, api: ABCAPI) -> bool:
        pass

    @abstractmethod
    def load(self, external: "ABCDispatch"):
        pass

    @abstractmethod
    def mount(self, view_t: typing.Type["ABCView"], name: str):
        pass



from telegrinder.model import Model
from telegrinder.result import Result
from telegrinder.api.error import APIError
import msgspec


class APIResponse(Model):
    ok: bool
    result: msgspec.Raw = b""
    error_code: int = 0
    description: str = ""

    def to_result(self) -> Result[msgspec.Raw, APIError]:
        if self.ok:
            return Result(True, value=self.result)
        return Result(False, error=APIError(self.error_code, self.description))



from .base import KeyboardSetBase, KeyboardSetError
from telegrinder.tools.keyboard import Keyboard, InlineKeyboard
import typing
import re
import os
import yaml


class KeyboardSetYAML(KeyboardSetBase):
    __config__: str

    @classmethod
    def load(cls) -> None:
        config_path = getattr(cls, "__config__", "keyboards.yaml")
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"Config file for {cls.__name__} is undefined")

        config = yaml.load(open(config_path, "r", encoding="utf-8"), yaml.Loader)
        for name, hint in typing.get_type_hints(cls).items():
            g = re.match(r"(?:kb_|keyboard_)(.+)", name.lower())
            if not g:
                continue

            short_name = g.group(1)
            if short_name not in config:
                raise KeyboardSetError(
                    f"Keyboard {short_name!r} is undefined in config"
                )

            kb_config = config[short_name]

            if (
                not isinstance(kb_config, dict)
                or "buttons" not in kb_config
                or not isinstance(kb_config["buttons"], list)
            ):
                raise KeyboardSetError(
                    "Keyboard should be dict with field buttons which must be a list, "
                    "check documentation"
                )

            buttons = kb_config.pop("buttons")
            new_keyboard: typing.Union[Keyboard, InlineKeyboard] = hint(**kb_config)

            for button in buttons:
                if not button:
                    new_keyboard.row()
                    continue
                if "text" not in button:
                    raise KeyboardSetError("text is required")
                new_keyboard.add(new_keyboard.BUTTON(**button))

            setattr(cls, name, new_keyboard)



import ssl
import typing

import aiohttp

from telegrinder.client.abc import ABCClient
from aiohttp import ClientSession, TCPConnector
from telegrinder.modules import json, JSONModule
import certifi

if typing.TYPE_CHECKING:
    from aiohttp import ClientResponse


class AiohttpClient(ABCClient):
    def __init__(
        self,
        session: typing.Optional[ClientSession] = None,
        json_processing_module: typing.Optional[JSONModule] = None,
        timeout: typing.Optional[aiohttp.ClientTimeout] = None,
        **session_params
    ):
        self.session = session
        self.json_processing_module = json_processing_module or json
        self.session_params = session_params
        self.timeout = timeout or aiohttp.ClientTimeout(total=0)

    async def request_raw(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> "ClientResponse":
        if not self.session:
            self.session = ClientSession(
                connector=TCPConnector(
                    ssl=ssl.create_default_context(cafile=certifi.where())
                ),
                json_serialize=self.json_processing_module.dumps,
                **self.session_params,
            )
        async with self.session.request(
            url=url, method=method, data=data, timeout=self.timeout, **kwargs
        ) as response:
            await response.read()
            return response

    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> dict:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.json(
            encoding="utf-8", loads=self.json_processing_module.loads, content_type=None
        )

    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: typing.Union[dict, aiohttp.FormData, None] = None,
        **kwargs
    ) -> str:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.text(encoding="utf-8")

    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: typing.Union[dict, aiohttp.FormData, None] = None,
        **kwargs
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        if response._body is None:
            await response.read()
        return response._body

    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return response._body

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

    @classmethod
    def get_form(cls, data: dict) -> aiohttp.formdata.FormData:
        form = aiohttp.formdata.FormData(quote_fields=False)
        for k, v in data.items():
            params = {}
            if isinstance(v, tuple):
                params["filename"], v = v[0], v[1]
            else:
                v = str(v)
            form.add_field(k, v, **params)
        return form

    def __del__(self):
        if self.session and not self.session.closed:
            if self.session._connector is not None and self.session._connector_owner:
                self.session._connector.close()
            self.session._connector = None



from .abc import Message, patcher
from .text import ABCTextMessageRule
import typing
import vbml

PatternLike = typing.Union[str, vbml.Pattern]


def check_string(
    patterns: typing.Union[PatternLike, typing.List[PatternLike]], s: str, ctx: dict
) -> bool:
    if s is None:
        return False
    for pattern in patterns:
        response = patcher.check(pattern, s)
        if response in (False, None):
            continue
        elif isinstance(response, dict):
            ctx.update(response)
        return True
    return False


class Markup(ABCTextMessageRule):
    def __init__(self, patterns: typing.Union[PatternLike, typing.List[PatternLike]]):
        if not isinstance(patterns, list):
            patterns = [patterns]
        self.patterns = [
            vbml.Pattern(pattern) if isinstance(pattern, str) else pattern
            for pattern in patterns
        ]

    async def check(self, message: Message, ctx: dict) -> bool:
        return check_string(self.patterns, message.text, ctx)



from .abc import Message
from .text import ABCTextMessageRule
import difflib
import typing


class FuzzyText(ABCTextMessageRule):
    def __init__(
        self, texts: typing.Union[str, typing.List[str]], min_ratio: float = 0.7
    ):
        if isinstance(texts, str):
            texts = [texts]
        self.texts = texts
        self.min_ratio = min_ratio

    async def check(self, message: Message, ctx: dict) -> bool:
        match = max(
            difflib.SequenceMatcher(a=message.text, b=text).ratio()
            for text in self.texts
        )
        if match < self.min_ratio:
            return False
        ctx["fuzzy_ratio"] = match
        return True



from .abc import ABCFormatter
import typing
from telegrinder.tools.parse_mode import ParseMode

QUOT_MARK = '"'


def wrap_tag(
    tag_name: str, content: str, data: typing.Optional[dict] = None
) -> "HTMLFormatter":
    s = "<" + tag_name
    if data:
        for k, v in data.items():
            s += f" {k}={v}"
    s += ">" + content + "</" + tag_name + ">"
    return HTMLFormatter(s)


class HTMLFormatter(ABCFormatter):
    PARSE_MODE = ParseMode.HTML

    def escape(self) -> "HTMLFormatter":
        return HTMLFormatter(
            self.replace("&", "&amp;").replace(">", "&gt;").replace("<", "&lt;")
        )

    def bold(self) -> "HTMLFormatter":
        return wrap_tag("b", self)

    def italic(self) -> "HTMLFormatter":
        return wrap_tag("i", self)

    def underline(self) -> "HTMLFormatter":
        return wrap_tag("u", self)

    def strike(self) -> "HTMLFormatter":
        return wrap_tag("s", self)

    def spoiler(self) -> "HTMLFormatter":
        return wrap_tag("tg-spoiler", self)

    def link(self, href: str) -> "HTMLFormatter":
        return wrap_tag(
            "a",
            self.escape(),
            data={"href": QUOT_MARK + HTMLFormatter(href).escape() + QUOT_MARK},
        )

    def code_block(self) -> "HTMLFormatter":
        return wrap_tag("pre", self.escape())

    def code_block_with_lang(self, lang: str) -> "HTMLFormatter":
        return wrap_tag(
            "pre", wrap_tag("code", self.escape(), data={"class": f"language-{lang}"})
        )

    def code_inline(self) -> "HTMLFormatter":
        return wrap_tag("code", self.escape())



import msgspec

from .abc import ABCAPI, APIError, Token
import typing
from telegrinder.result import Result
from telegrinder.client import ABCClient, AiohttpClient
from telegrinder.types.methods import APIMethods
from telegrinder.model import convert
from telegrinder.api.response import APIResponse


def compose_data(client: ABCClient, data: dict) -> typing.Any:
    data = {k: convert(v) for k, v in data.items()}
    if any(isinstance(v, tuple) for v in data.values()):
        data = client.get_form(data)
    return data


class API(ABCAPI, APIMethods):
    API_URL = "https://api.telegram.org/"

    def __init__(self, token: Token, http: typing.Optional[ABCClient] = None):
        self.token = token
        self.http = http or AiohttpClient()
        super().__init__(self)

    @property
    def id(self) -> int:
        return int(self.token.split(":")[0])

    @property
    def request_url(self) -> str:
        return self.API_URL + f"bot{self.token}/"

    async def request(
        self,
        method: str,
        data: typing.Optional[dict] = None,
    ) -> Result[typing.Union[dict, list, bool], APIError]:
        data = compose_data(self.http, data)
        response = await self.http.request_json(self.request_url + method, data=data)
        if response.get("ok"):
            assert "result" in response
            return Result(True, value=response["result"])

        code, msg = response.get("error_code"), response.get("description")
        return Result(False, error=APIError(code, msg))

    async def request_raw(
        self,
        method: str,
        data: typing.Optional[dict] = None,
    ) -> Result[msgspec.Raw, APIError]:
        data = compose_data(self.http, data)
        response_bytes = await self.http.request_bytes(
            self.request_url + method, data=data
        )
        response_skeleton: APIResponse = msgspec.json.decode(
            response_bytes, type=APIResponse
        )
        return response_skeleton.to_result()



from telegrinder.types import InlineQuery, User
from telegrinder.api import API, APIError
from telegrinder.result import Result
import typing


class InlineQueryCute(InlineQuery):
    api: API

    @property
    def from_user(self) -> User:
        return self.from_

    @property
    def ctx_api(self) -> API:
        return self.api

    async def answer(
        self,
        results: typing.Optional[list] = None,
        cache_time: typing.Optional[int] = None,
        is_personal: typing.Optional[bool] = None,
        next_offset: typing.Optional[str] = None,
        switch_pm_text: typing.Optional[str] = None,
        switch_pm_parameter: typing.Optional[str] = None,
    ) -> Result[bool, APIError]:
        return await self.ctx_api.answer_inline_query(
            self.id,
            results=results,
            cache_time=cache_time,
            is_personal=is_personal,
            next_offset=next_offset,
            switch_pm_text=switch_pm_text,
            switch_pm_parameter=switch_pm_parameter,
        )



from .abc import ABCView
from telegrinder.bot.dispatch.handler import ABCHandler, FuncHandler
from telegrinder.bot.dispatch.waiter import Waiter
from telegrinder.bot.dispatch.middleware import ABCMiddleware
from telegrinder.bot.rules import ABCRule
from telegrinder.bot.cute_types import MessageCute
from telegrinder.api.abc import ABCAPI
from telegrinder.bot.dispatch.waiter import WithWaiter, DefaultWaiterHandler
from telegrinder.bot.dispatch.process import process_waiters, process_inner
from telegrinder.types import Update
import typing


class MessageView(ABCView, WithWaiter[int, MessageCute]):
    def __init__(self):
        self.handlers: typing.List[ABCHandler[MessageCute]] = []
        self.middlewares: typing.List[ABCMiddleware[MessageCute]] = []
        self.short_waiters: typing.Dict[int, Waiter] = {}

    def __call__(self, *rules: ABCRule, is_blocking: bool = True):
        def wrapper(func: typing.Callable[..., typing.Coroutine]):
            self.handlers.append(
                FuncHandler(func, list(rules), is_blocking, dataclass=None)
            )
            return func

        return wrapper

    def load(self, external: "MessageView"):
        self.handlers.extend(external.handlers)
        self.middlewares.extend(external.middlewares)
        external.short_waiters = self.short_waiters

    async def check(self, event: Update) -> bool:
        return bool(event.message)

    async def process(self, event: Update, api: ABCAPI):
        msg = MessageCute(**event.message.to_dict(), api=api)

        if await process_waiters(
            self.short_waiters, msg.chat.id, msg, event, msg.answer
        ):
            return

        return await process_inner(msg, event, self.middlewares, self.handlers)

    async def wait_for_message(
        self,
        chat_id: int,
        *rules: ABCRule,
        default: typing.Optional[typing.Union[DefaultWaiterHandler, str]] = None
    ) -> typing.Tuple["MessageCute", dict]:
        return await self.wait_for_answer(chat_id, *rules, default=default)



import msgspec
from telegrinder.result import Result
from telegrinder.modules import json
from msgspec import Raw
import typing

if typing.TYPE_CHECKING:
    from telegrinder.api.error import APIError

T = typing.TypeVar("T")
encoder = msgspec.json.Encoder()


def full_result(
    result: Result[msgspec.Raw, "APIError"], full_t: typing.Type[T]
) -> Result[T, "APIError"]:
    if not result.is_ok:
        return result
    return Result(True, value=msgspec.json.decode(result.value, type=full_t))


def convert(d: typing.Any) -> typing.Any:
    if isinstance(d, Model):
        return msgspec.json.encode(d).decode()
    elif isinstance(d, dict):
        return {k: convert(v) for k, v in d.items() if v is not None}
    elif isinstance(d, list):
        return json.dumps(d)
    return d


model_config = {"rename": {"from_": "from"}, "omit_defaults": True}


class Model(msgspec.Struct, **model_config):
    _dict_cached: typing.Optional[dict] = None

    def to_dict(self) -> dict:
        if self._dict_cached is not None:
            return self._dict_cached
        self._dict_cached = {k: getattr(self, k) for k in self.__struct_fields__}
        return self._dict_cached


def get_params(params: dict) -> dict:
    return {
        k: v for k, v in (
            *params.items(),
            *params.pop("other").items()
        )
        if k != "self"
        and v is not None
    }


__all__ = (
    "convert",
    "model_config",
    "encoder",
    "full_result",
    "msgspec",
    "Model",
    "Raw",
    "get_params",
)



from .abc import ABCRule, ABCMessageRule, AndRule, OrRule
from .callback_data import (
    CallbackDataEq,
    CallbackDataJsonEq,
    CallbackDataJsonModel,
    CallbackDataMarkup,
)
from .func import FuncRule
from .is_from import IsPrivate, IsChat
from .markup import Markup
from .regex import Regex
from .text import Text, HasText, ABCTextMessageRule
from .fuzzy import FuzzyText
from .integer import Integer, IntegerInRange



from .abc import ABCRule, EventScheme, T
from telegrinder.types import Update
import typing


class FuncRule(ABCRule, typing.Generic[T]):
    def __init__(
        self,
        func: typing.Callable[[T, dict], bool],
        event_scheme: typing.Optional[
            typing.Union[EventScheme, typing.Tuple[str, typing.Type[T]]]
        ] = None,
    ):
        self.func = func
        if isinstance(event_scheme, tuple):
            event_scheme = EventScheme(*event_scheme)
        self.event_scheme = event_scheme

    async def check(self, event: Update, ctx: dict) -> bool:
        if self.event_scheme:
            if self.event_scheme.name not in event:
                return False
            event = self.event_scheme.dataclass(
                **getattr(event, self.event_scheme.name).to_dict()
            )
        return self.func(event, ctx)



from abc import ABC
import typing

T = typing.TypeVar("T")


class ABCMiddleware(ABC, typing.Generic[T]):
    async def pre(self, event: T, ctx: dict) -> bool:
        ...

    async def post(self, event: T, responses: typing.List[typing.Any], ctx: dict):
        ...



from abc import ABC, abstractmethod
from telegrinder.types import Update
from telegrinder.api.abc import ABCAPI
import typing

T = typing.TypeVar("T")


class ABCHandler(ABC, typing.Generic[T]):
    is_blocking: bool
    ctx: dict

    @abstractmethod
    async def run(self, event: T) -> typing.Any:
        pass

    @abstractmethod
    async def check(self, api: ABCAPI, event: Update) -> bool:
        pass



import logging
import typing

from choicelib import choice_in_order
from typing_extensions import Protocol


class JSONModule(Protocol):
    def loads(self, s: str) -> typing.Union[dict, list]:
        ...

    def dumps(self, o: typing.Union[dict, list]) -> str:
        ...


json: JSONModule = choice_in_order(
    ["ujson", "hyperjson", "orjson"], do_import=True, default="json"
)

logging_module = choice_in_order(["loguru"], default="logging")

if logging_module == "loguru":
    import os
    import sys

    if not os.environ.get("LOGURU_AUTOINIT"):
        os.environ["LOGURU_AUTOINIT"] = "0"
    from loguru import logger  # type: ignore

    if not logger._core.handlers:  # type: ignore
        log_format = (
            "<level>{level: <8}</level> | "
            "{time:YYYY-MM-DD HH:mm:ss} | "
            "{name}:{function}:{line} > <level>{message}</level>"
        )
        logger.add(sys.stderr, format=log_format, enqueue=True, colorize=True)

elif logging_module == "logging":
    """
    This is a workaround for lazy formatting with {} in logging.
    About:
    https://docs.python.org/3/howto/logging-cookbook.html#use-of-alternative-formatting-styles
    """
    import inspect
    import logging

    class LogMessage:
        def __init__(self, fmt, args, kwargs):
            self.fmt = fmt
            self.args = args
            self.kwargs = kwargs

        def __str__(self):
            return self.fmt.format(*self.args)

    class StyleAdapter(logging.LoggerAdapter):
        def __init__(self, logger, extra=None):
            super().__init__(logger, extra or {})

        def log(self, level, msg, *args, **kwargs):
            if self.isEnabledFor(level):
                msg, args, kwargs = self.proc(msg, args, kwargs)
                self.logger._log(level, msg, args, **kwargs)

        def proc(self, msg, args, kwargs):
            log_kwargs = {
                key: kwargs[key]
                for key in inspect.getfullargspec(self.logger._log).args[1:]
                if key in kwargs
            }
            if isinstance(msg, str):
                msg = LogMessage(msg, args, kwargs)
                args = ()
            return msg, args, log_kwargs

    logger = StyleAdapter(logging.getLogger("telegrinder"))  # type: ignore



import re
import typing

from .abc import Message
from .text import ABCTextMessageRule

PatternLike = typing.Union[str, typing.Pattern]


class Regex(ABCTextMessageRule):
    def __init__(self, regexp: typing.Union[PatternLike, typing.List[PatternLike]]):
        if isinstance(regexp, re.Pattern):
            regexp = [regexp]
        elif isinstance(regexp, str):
            regexp = [re.compile(regexp)]
        else:
            regexp = [
                re.compile(regexp) if isinstance(regexp, str) else regexp
                for regexp in regexp
            ]

        self.regexp = regexp

    async def check(self, message: Message, ctx: dict) -> bool:
        for regexp in self.regexp:
            match = re.match(regexp, message.text)
            if match:
                ctx.update({"match": match.groups()})
                return True
        return False
